import importlib
import itertools
from shlex import shlex
import sys
# from time import sleep

from chart import Chart
from symbols import Symbol
from log import log, set_log_level, DEBUG
from rule import Rule
from rules import Rules
from vertical import XmlTag


class ChartParser:
    def __init__(self, rules=Rules(), rule_files=[], argv=[]):
        self.options = dict(
            debug=False,  # print under the hood information
            chart=False,
            edges=False,  # show each edge as it is handled and combined
            # original only, no dependencies yet (--dependency/-d)
            vertical=False,  # TODO: accept file names?
            phrasal_trees=False,  # constituency_trees
            dot=False,  # trees in the DOT format
            interactive=False,  # stop after each processed sentence
            )
        if rules:
            self.load_rules()
        else:
            if not rule_files and argv:
                rule_files, options = self.read_cmdline(argv[1:])
                self.options.update(options)

            if not rule_files:
                rule_files = ['cz']
            self.load_rules_from_files(rule_files)

        if self.options['debug']:
            set_log_level(DEBUG)
        self.print_rules()

    def parse(self, tokens, root='CLAUSE', opening=None, closing=None):
        self.tokens = tokens

        self._add_parsed_attibutes_to_tokens()
        for token in tokens:
            log.info(Symbol(token))

        # když mezi nima není sloveso nebo „lze“ k6, spustit parse(root='NP')
        # teď to není potřeba, protože CLAUSE predikuje i NP
        # ale bude se to hodit na vernacular a optional_punctuation
        if False:
            log.info('No verb found, parsing as NP instead')

        self.root = root
        self.known_edges = set()
        self.prepare_agenda()
        self.chart = Chart()

        while self.agenda:
            agenda_edge = self.agenda.popleft()
            for new_edge in self.propose_edges(agenda_edge):
                self.agenda.append(new_edge)
            self.chart.append(agenda_edge)

        if self.options['chart']:
            self.chart.print()
        length = len(self.tokens)
        self.successful_edges = [
            edge for edge in self.chart.closed_edges
            if edge.begin == 0 and edge.end == length]
        if not self.successful_edges:
            log.error('\nNo edge covering the whole input!\n')
            # TODO: vytisknout nejdelší hrany (nejzazší edge.end)
        for edge in self.successful_edges:
            log.info('\nPassed!      %s\n', edge)
            if self.options['phrasal_trees']:
                self.print_graph(edge)
                log.info('\n')
            if self.options['dot']:
                self.print_dot(edge)
        if opening is not None:
            opening['passed'] = 'yes' if self.successful_edges else 'no'
        if self.options['vertical']:
            self.print_vertical(tokens, opening, closing)
        # if self.interactive:
        #     input()

    def print_rules(self):
        for rule in self.rules:
            log.debug('rule         %s', rule)
        log.info('\n')

    @staticmethod
    def extract_attributes_from_tag(tag):
        if ',' in tag:
            tags = tag.split(',')
            if '' not in tags:
                return ChartParser.merge_ambiguous_tags(tags)
        elif len(tag) % 2 != 0:
            raise Exception('Malformed tag "%s"', tag)
        return ((attr, val) for (attr, val) in grouper(tag, 2))

    @staticmethod
    def merge_ambiguous_tags(tags):
        merged = {}
        for attr, val in itertools.chain.from_iterable(
                ChartParser.extract_attributes_from_tag(tag) for tag in tags):
            if attr not in merged:
                merged[attr] = val
            elif isinstance(merged[attr], str):
                if val != merged[attr]:
                    merged[attr] = [merged[attr], val]
            elif val not in merged[attr]:
                merged[attr].append(val)
        return merged

    def _add_parsed_attibutes_to_tokens(self):
        for token in self.tokens:
            if 'tag' in token:
                token.update(
                    self.extract_attributes_from_tag(token['tag']))

    def parse_from_vertical(self, lines=sys.stdin, root='CLAUSE'):
        tokens = []
        opening = None
        phrase = None

        for line in lines:
            line = line.strip()
            if line.startswith('<'):
                tag = XmlTag.parse(line)
                if tag.tag in ('s', 'head'):
                    closing = tag if tag.closing else None
                    if closing is not None or tokens:
                        self.parse(tokens, opening=opening, closing=closing)
                        tokens = []
                        # sleep(0.1)
                    opening = tag if tag.opening else None
                    continue

                if tag.tag == 'g':  # and empty
                    if tokens:
                        # the next token is “glued” to this one
                        tokens[-1].trailing_whitespace = False
                elif tag.tag == 'phr':  # MWE
                    if tag.opening:
                        phrase = Symbol(word=[], lemma=tag.pop('l', None),
                                        tag=tag.pop('t', None), mwe=True)
                    elif phrase:
                        phrase['word'] = ' '.join(phrase['word'])
                        tokens.append(phrase)
                        phrase = None
                    else:
                        log.warning('unexpected </phr>!')
                elif tag.tag in ('doc', 'p'):
                    pass  # zatím neřeším (parseru to může být skoro jedno)
                elif tag.tag == 'table':
                    # přeskakovat někdy celý obsah?
                    pass  # table = True
                elif tag.tag == 'sign':
                    pass  # zajímavé: Josef Juřeník, přednosta stanice
            elif line:
                if phrase is not None:
                    phrase['word'].append(line)
                else:
                    try:
                        word, lemma, tag = line.split()
                    except ValueError:
                        word = lemma = line
                        if line in ',.():?':
                            tag = 'kI'  # cesta do pekel :-D
                        elif line[0] in '0123456789':  # TODO: vylepšit heuristiku
                            log.warning('line with a number, correct? ' + line)
                            tag = 'k4'
                        else:  # 31. března
                            log.warning('unexpected line ' + line)
                            tag = 'k?'
                    token = Symbol(word=word, lemma=lemma, tag=tag, line=line)
                    tokens.append(token)
        if tokens:
            self.parse(tokens, root=root)

    def load_rules_from_files(self, rule_files):
        self.rules = Rules()
        for rules_filename in rule_files:
            rules_file = importlib.import_module('rules.' + rules_filename)
            self.load_rules(rules_file.rules)

    def load_rules(self, rules):
        for rule in rules:
            if not isinstance(rule, Rule):
                rule = Rule(rule)
            self.rules.append(rule)
            for epsilon_rule in rule.epsilon_rules:
                self.rules.append(epsilon_rule)

    def print_graph(self, edge, level=0):
        log.info('{0}{1}'.format('   '*level, edge))
        level += 1
        for child in edge:
            if child.edge is True:
                if len(edge) > 1:
                    log.info('{0}{1}'.format('   '*level, child))
            elif child.edge and child.length:
                self.print_graph(child.edge, level)

    def print_dot(self, edge, parent_indexes='0'):
        if parent_indexes == '0':
            print('graph {')
        print('    {} [label="{}"];'.format(parent_indexes, edge.left))
        for index, child in enumerate(edge):
            if child.edge is True:
                print('    {}{} [label="{}"];'.format(
                    parent_indexes, index, child))
            elif child.edge and child.length:
                self.print_dot(child.edge, parent_indexes + str(index))
            else:
                continue
            print('    {0} -- {0}{1};'.format(parent_indexes, index))
        if parent_indexes == '0':
            print('}')

    @staticmethod
    def print_vertical(tokens, opening, closing):
        if opening:
            print(opening)
        for token in tokens:
            print(token.vertical())
        if closing is not None:
            print(closing)

    @staticmethod
    def read_cmdline(argv):
        rules = []
        options = {}
        expect_rules = False
        for arg in argv:
            if arg == '--rules':
                expect_rules = True
                continue
            if expect_rules:
                rules.append(arg)
                expect_rules = False
            elif arg == '--debug':
                options['debug'] = True
            elif arg == '--edges':
                options['edges'] = True
            elif arg == '--vertical':
                options['vertical'] = True
            elif arg == '--chart':
                options['chart'] = True
            elif arg == '--trees':
                options['phrasal_trees'] = True
            elif arg == '--dot':
                options['dot'] = True
            # elif arg == '--interactive':
            #     options['interactive'] = True
            else:
                log.warning('Unknown parameter: ' + arg)
        return rules, options


# copied from oVirt/VDSM, which in turn used
# http://docs.python.org/2.6/library/itertools.html?highlight=grouper#recipes
def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return itertools.zip_longest(fillvalue=fillvalue, *args)
